
--
--  Gossipmonger
--    by Tuhljin
--

local THIS_VERSION = GetAddOnMetadata("Gossipmonger", "Version");
local THIS_TITLE = GetAddOnMetadata("Gossipmonger", "Title")

local L = GOSSIPMONGER_STRINGS

local CYAN = "|cff00ffff"
local YELLOW = "|cffffff00"
local YELLOW_LT = "|cffffff99"
local R, G, B = 0.494, 0.729, 1
--local SCTColors

local minimumGossipHeight, maximumGossipHeight = 150, 414
local incGossipHeight = 119

local ChatIcon = "|TInterface\\GossipFrame\\GossipGossipIcon:0:0:2:-2|t"

Gossipmonger = {};


--Gossipmonger_Debug = true
--Gossipmonger.DEBUG_USETITLED = true


local CurrentNPC, CurrentNPC_gossip
local SafeMirrorOptionsClick
local GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip  -- For the GossipFrame_OnEvent hook
local IgnoreQuestFinished, FollowupFunction_filtername
local typeReplacement

local SPIRIT_HEALER_KEYS = { "XP_LOSS", "XP_LOSS_NO_DURABILITY", "XP_LOSS_NO_SICKNESS_NO_DURABILITY", "XP_LOSS_NO_SICKNESS" }


local function chatprint(msg)
  DEFAULT_CHAT_FRAME:AddMessage(msg, R, G, B);
end


local function getMobID(unit)
  local guid = UnitGUID(unit)
  if (not guid) then  return;  end
  local unitType, _, _, _, _, id = ("-"):split(guid)
  if (unitType == "Creature") then
    return tonumber(id)
  end
end


local informMsg, informTimer

local function timedInform()
  if (informMsg) then
    if (not Gossipmonger_MirrorFrame:IsVisible()) then
      chatprint(informMsg)
    end
    informMsg = nil
  end
end

local function doInform(msg)
  if (Gossipmonger_Settings.informOnSkip_nobubble) then
    informMsg = msg
	if (informTimer and not informTimer._cancelled) then  informTimer:Cancel();  end
	-- 1 second should be enough most of the time, from my experience, but lag could mean the message goes off when ideally it shouldn't. Oh well.
	informTimer = C_Timer.NewTimer(1, timedInform)  -- Note that you needn't use an integer. Could go with 0.5 seconds, for instance.
  else
    chatprint(msg)
  end
end

local function checkInform(override, validFollowup, gossipType, name)
	--local inform_SCT, inform_log = Gossipmonger_Settings.informOnSkip_SCT and SCT, Gossipmonger_Settings.informOnSkip_log
	local inform_log = Gossipmonger_Settings.informOnSkip_log
	--if (override ~= false and (inform_SCT or inform_log)) then
	if (override ~= false and not validFollowup and (inform_log)) then
	  local named, inform = L.SKIPPINGTEXT_NAME:format(name)
	  if (not gossipType or ((Gossipmonger_Settings.alwaysInformWithName or strlen(name) <= 20) and not Gossipmonger.DEBUG_USETITLED)) then
		inform = named
	  else
		if (not typeReplacement) then
		  -- Creating lookup table
		  typeReplacement = {}
		  local arr = { strsplit("\n", L.NPCTYPE_REPLACE) }
		  for k,v in ipairs(arr) do
			local t, r = strsplit("=", v, 2)
			if (t and r and t ~= "" and r ~= "") then  typeReplacement[t] = r;  end
		  end
		end
		local titled = L.SKIPPINGTEXT_TYPE:format(typeReplacement[gossipType] and typeReplacement[gossipType] or gossipType)
		inform = strlen(named) <= strlen(titled) and named or titled
		if (Gossipmonger.DEBUG_USETITLED) then  inform = titled;  end
	  end
	  --if (inform_SCT) then
	  --  SCTColors = SCTColors or { r = R, g = G, b = B }
	  --  SCT:DisplayMessage(ChatIcon..inform, SCTColors)
	  --end
	  if (inform_log) then
		doInform(ChatIcon.." "..CYAN..inform)
	  end
	end
end


local TQD, preTQD

local function prepGossipOptionDisplay(id, baseButton, selected, QuestGreeting)
  local name = "Gossipmonger_OptionDisplay"..id
  local titleButton = _G[name]
  if (not titleButton) then
    titleButton = CreateFrame("Button", name, Gossipmonger_MirrorFrame_GreetingScrollFrameScrollChild, "Gossipmonger_OptionDisplay")
    if (id == 1) then
      titleButton:SetPoint("TOPLEFT", Gossipmonger_MirrorFrame_GreetingText, "BOTTOMLEFT", -10, -20)
    else
      local prevButton = _G["Gossipmonger_OptionDisplay"..(id - 1)]
      titleButton:SetPoint("TOPLEFT", prevButton, "BOTTOMLEFT", 0, -3)
    end
  end

  titleButton.baseButton = baseButton
  local baseIcon = QuestGreeting and _G[baseButton.."QuestIcon"] or _G[baseButton.."GossipIcon"]
  titleButton.QuestGreeting = QuestGreeting
  baseButton = _G[baseButton]

  if (baseButton:IsShown()) then
    titleButton:Show()

    if (TRIVIAL_QUEST_DISPLAY ~= preTQD) then
      TQD = TRIVIAL_QUEST_DISPLAY:gsub("%(","%%(")
      TQD = TQD:gsub("%)","%%)")
      TQD = ("^"..TQD.."$"):gsub("%%s", "(.+)")
      preTQD = TRIVIAL_QUEST_DISPLAY
    end
    local NQD = ("^"..NORMAL_QUEST_DISPLAY.."$"):gsub("%%s", "(.+)")
    local text = baseButton:GetText()
    local _, baseTitle, taggedTitle
    _, _, baseTitle = text:find(TQD)
    _, _, taggedTitle = text:find(NQD)
    if (baseTitle) then  -- Low level quest
      titleButton.isLowLevel = true
    elseif (taggedTitle) then  -- Normal quest
      titleButton.isLowLevel = nil
      baseTitle = taggedTitle
    else
      baseTitle, taggedTitle = text, text
    end
    titleButton.title = baseTitle

    local title = selected and YELLOW .. taggedTitle .. "   " .. YELLOW_LT .. L.SELECTEDOPTION or
                  YELLOW .. taggedTitle
    titleButton:SetText(title);
    GossipResize(titleButton);   -- Calls a Blizzard function that currently simply does this:  titleButton:SetHeight( titleButton:GetTextHeight() + 2);
    titleButton:SetID( baseButton:GetID() )
    if (baseButton.type) then
      titleButton.type = baseButton.type  -- Set to "Available", "Active", or "Gossip"
    elseif (baseButton.isActive == 1) then
      titleButton.type = "Active"
    elseif (baseButton.isActive == 0) then
      titleButton.type = "Available"
    else
      titleButton.type = nil  -- failsafe
    end
    local icon = _G[name.."GossipIcon"]
    icon:SetTexture( baseIcon:GetTexture() )
    icon:SetVertexColor( baseIcon:GetVertexColor() )
    titleButton.selected = selected
    return baseTitle, titleButton:GetHeight()
  else
    titleButton.selected = nil
    titleButton:Hide()
  end
end

local orig_GossipFrameOptionsUpdate = GossipFrameOptionsUpdate
local function safe_GossipFrameOptionsUpdate(...)
	local num = select("#", ...)
	--print("safe_GossipFrameOptionsUpdate",select("#", ...),...)
	if (num == 1 or (num == 2 and select(2, ...) == nil)) then -- If there are options, we'd expect 2 strings or more, but in odd cases (e.g. talk to Class Hall Upgrades NPC), we get only one, which results in an error in Blizzard's Lua (since they didn't expect to have to show an icon for a gossip option that gets skipped over anyway).
		return orig_GossipFrameOptionsUpdate(select(1, ...), "gossip") -- Use gossip chat bubble icon by default.
	end
	return orig_GossipFrameOptionsUpdate(...)
end
GossipFrameOptionsUpdate = safe_GossipFrameOptionsUpdate

local function showGossipMirror(delay, name, text, selected, gossipType, noUpdate, watchQuest, numQuests, QuestGreeting)
  SafeMirrorOptionsClick = nil
  Gossipmonger_MirrorFrame_GreetingText:SetText(text);
  if (not noUpdate) then  -- noUpdate should be true when ShowGossipFrame already handled this
    GossipFrameUpdate() -- Update standard buttons; we'll grab the data from them rather than reinventing the
                        -- wheel. (It's hopefully more forward compatible this way, as well.)
  end
  -- Show used buttons:
  local optionsHeight, mirrorbtnIndex, h = -3, 1
  local buttonName = QuestGreeting and "QuestTitleButton" or "GossipTitleButton"
  local selectedText
  for i = 1, (QuestGreeting and numQuests or GossipFrame.buttonIndex - 1) do
  -- When using GossipFrame.buttonIndex, the actual number of buttons displayed is one less than given.
    local thisSelected = selected == mirrorbtnIndex
	local thisText
    thisText, h = prepGossipOptionDisplay(mirrorbtnIndex, buttonName..i, thisSelected, QuestGreeting)
	if (thisSelected) then  selectedText = thisText;  end
    if (h) then
      optionsHeight = optionsHeight + h + 3
      mirrorbtnIndex = mirrorbtnIndex + 1
    end
  end

  if (not selectedText) then -- Failed to find the selected option. Let QuestHanger do its search. (Failsafe for when selected variable doesn't match anything. Not sure exact circumstance it's needed, but it was for a Wanted Board object at Refuge Pointe that didn't have a proper NPC associated with it.)
    Gossipmonger_MirrorFrame.selected = nil
  else
    Gossipmonger_MirrorFrame.selected = selected
  end
  Gossipmonger_MirrorFrame.numOptions = mirrorbtnIndex - 1
  Gossipmonger_MirrorFrame.numQuests = numQuests or GetNumGossipAvailableQuests() + GetNumGossipActiveQuests()

  -- Hide the rest:
  local titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  while (titleButton) do
    titleButton:Hide()
    mirrorbtnIndex = mirrorbtnIndex + 1
    titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  end

  h = Gossipmonger_MirrorFrame_GreetingText:GetHeight() or 0
  local h2 = h + optionsHeight + incGossipHeight
  if (h2 < minimumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(minimumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  elseif (h <= Gossipmonger_MirrorFrame_GreetingScrollFrame:GetHeight() and h2 <= maximumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(h2)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  else
    Gossipmonger_MirrorFrame:SetHeight(maximumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Show()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel()
  end

  CurrentNPC, CurrentNPC_gossip = name, gossipType

  Gossipmonger_MirrorFrame_NpcNameText:SetText( CurrentNPC or CYAN..THIS_TITLE )
  if ( UnitExists("npc") ) then
    Gossipmonger_MirrorFrame_NpcNameText:SetTextColor(GameTooltip_UnitColor("npc"))
    SetPortraitTexture(Gossipmonger_MirrorFrame_Portrait, "npc");
    Gossipmonger_MirrorFrame.NPCName = CurrentNPC
  else
    Gossipmonger_MirrorFrame_Portrait:SetTexture("Interface\\QuestFrame\\UI-QuestLog-BookIcon");
    Gossipmonger_MirrorFrame.NPCName = CurrentNPC --nil
  end

  if (not delay) then
  -- When the Gossip Bubble might be put in a special location, we have to act like delay option is enabled even when it's not.
  -- Otherwise, if it turns out that it should be in that special location, it will appear in the wrong spot first and then move
  -- in a jarring fashion to the proper spot later.
  -- This is the case with the new (Legion) style flight map:
    if (gossipType == "taxi") then
	  --and not TaxiFrame_ShouldShowOldStyle() -- Unfortunately, can't use this because TaxiFrame_ShouldShowOldStyle() doesn't return the correct value unless the map is opened.
	  delay = true
  -- It's also the case with the NPC recruiter if you've already chosen a counter/trait (and while not the case if you haven't chosen that yet, we can't really tell until it's too late so we simply always use delay):
    elseif (gossipType == "chatbubble" and selectedText == L.SHOWBUBBLE_DELAY_FORCE_OPTION) then
	  delay = true
	end
  end

  Gossipmonger_MirrorFrame_GreetingScrollFrame:SetVerticalScroll(0);
  if (not delay) then
    ShowUIPanel(Gossipmonger_MirrorFrame)
	informMsg = nil
  end

  if (watchQuest) then
    if (watchQuest == true or Gossipmonger_Settings.showGossipMirrorOnQuestSelect) then
	-- Old comment preserved (for now) in case I need it later, but I'm not sure it's accurate or reflects what I want any more:
    -- watchQuest is set to 1 if the mirror should only automatically pops up if the settings tell it to do so,
    -- or true if it should always pop up (as is the case when a filter selects it). Either way, if the player
    -- clicked the vendor option, for example, instead of a quest, the mirror shouldn't pop up since it wasn't
    -- due to a filter selecting it.
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_DETAIL")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_PROGRESS")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_COMPLETE")
	  --Gossipmonger_MirrorFrame:RegisterEvent("QUEST_ACCEPTED") --Don't register it here. The way we do it, has to be after QUEST_DETAIL triggers or things break.
    end
  else
    Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_BINDER") -- Innkeeper
	Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_XP_LOSS") -- Spirit Healer
	Gossipmonger_MirrorFrame:RegisterEvent("AUCTION_HOUSE_SHOW")
	Gossipmonger_MirrorFrame:RegisterEvent("SHIPMENT_CRAFTER_INFO") --SHIPMENT_CRAFTER_OPENED happens a split second too soon
	Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUITMENT_NPC_OPENED") -- Window where follower candidates' counter/trait is chosen is shown
	Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUITMENT_READY") -- Window where resulting follower candidates are shown
	Gossipmonger_MirrorFrame:RegisterEvent("TRADE_SKILL_SHOW") -- For garrison profession crafting. A player could, if they timed it just right, trigger this by opening their own skill window, but it shouldn't really matter.
	Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_TALENT_NPC_OPENED") -- Order Hall talents
  end
  --[[
  Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_OPENED")
  Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_OPENED")
  Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_BINDER")
  Gossipmonger_MirrorFrame:RegisterEvent("AUCTION_HOUSE_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("SHIPMENT_CRAFTER_INFO") --SHIPMENT_CRAFTER_OPENED happens a split second too soon
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUITMENT_NPC_OPENED")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUITMENT_READY")
  Gossipmonger_MirrorFrame:RegisterEvent("TRADE_SKILL_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_TALENT_NPC_OPENED")
  --]]
end

local function BreakMirror() -- If you ever change this signature (adding arguments), make sure to change the "DETECT POPUP CLOSING" references to this to explicitly call it without arguments instead of just making it the OnHide (with self passed in, likely)
-- We're not just hiding the mirror. We're "breaking" it away from the current NPC. (There's a distinction because
-- there may be times we want to have the option of showing it again w/o the disassociation that goes on here.)
  CurrentNPC, CurrentNPC_gossip, SafeMirrorOptionsClick = nil, nil, nil
  IgnoreQuestFinished = nil
  Gossipmonger_MirrorFrame.NPCName = nil
  if (Gossipmonger_MirrorFrame:IsShown()) then
    HideUIPanel(Gossipmonger_MirrorFrame)
  end
  Gossipmonger_MirrorFrame:UnregisterAllEvents();
  informMsg = nil
end

local specialSetup

local function ShowMirror(specialPosition)
  if (not Gossipmonger_MirrorFrame:IsShown()) then
    if (specialPosition) then
	  if (not specialSetup) then
	    specialSetup = true
		Gossipmonger_MirrorFrame:SetClampedToScreen(true)
	  end
	  --Gossipmonger_MirrorFrame:SetPoint("TOPRIGHT", FlightMapFrame, "TOPLEFT", -10, 0)
	  --Gossipmonger_MirrorFrame:SetPoint("TOPLEFT", FlightMapFrame, "TOPLEFT", - Gossipmonger_MirrorFrame:GetWidth() - 10, -50)
	  Gossipmonger_MirrorFrame:SetPoint("TOPLEFT", specialPosition, "TOPLEFT", 115, Gossipmonger_MirrorFrame:GetHeight() + 5)
	  Gossipmonger_MirrorFrame:Show()
	else
      ShowUIPanel(Gossipmonger_MirrorFrame)
	end
  end
  SafeMirrorOptionsClick = true
  informMsg = nil

  Gossipmonger_MirrorFrame:UnregisterAllEvents();
  Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("QUEST_FINISHED")
  Gossipmonger_MirrorFrame:RegisterEvent("AUCTION_HOUSE_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("SHIPMENT_CRAFTER_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUITMENT_NPC_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_RECRUIT_FOLLOWER_RESULT")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_TRADESKILL_NPC_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("GARRISON_TALENT_NPC_CLOSED")
end

local function UpdateStandardFilters()
  if (Gossipmonger_Settings.Filter_WoWDefault) then
    Gossipmonger.EnableFilter(Gossipmonger.StandardFilters["Default"])
  else
    Gossipmonger.DisableFilter(Gossipmonger.StandardFilters["Default"])
  end
  if (Gossipmonger_Settings.Filter_Accuracy) then
    Gossipmonger.EnableFilter(Gossipmonger.StandardFilters["Accuracy"])
  else
    Gossipmonger.DisableFilter(Gossipmonger.StandardFilters["Accuracy"])
  end
  if (Gossipmonger_Settings.Filter_Mundane) then
    Gossipmonger.EnableFilter(Gossipmonger.StandardFilters["Mundane"])
  else
    Gossipmonger.DisableFilter(Gossipmonger.StandardFilters["Mundane"])
  end
end

local function Init()
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Gossipmonger.OnEvent);

  local panel, oldver = Gossipmonger.CreateOptions(THIS_TITLE, BreakMirror, UpdateStandardFilters)

  if (oldver and oldver ~= THIS_VERSION) then
    Gossipmonger_Settings.Version = THIS_VERSION
    for k,v in pairs(Gossipmonger.DefaultSettings) do
    -- Add options new to this version, set to default:
      if (Gossipmonger_Settings[k] == nil) then
        Gossipmonger_Settings[k] = v;
      end
    end
  end

  UpdateStandardFilters()

  do
	local xoffset = UIPanelWindows.Gossipmonger_MirrorFrame.xoffset

	local function resetOffset()
		SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", xoffset)
		if (Gossipmonger_MirrorFrame:IsVisible()) then
			HideUIPanel(Gossipmonger_MirrorFrame)
			ShowUIPanel(Gossipmonger_MirrorFrame)
		end
	end
	--/dump Gossipmonger_MirrorFrame:GetAttribute("UIPanelLayout-xoffset")

	-- Adjust position based on whether Quest NPC Model is shown:
	QuestNPCModel:HookScript("OnShow", function()
		SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", (xoffset or 0) + QuestNPCModel:GetWidth())
		if (Gossipmonger_MirrorFrame:IsVisible()) then
			HideUIPanel(Gossipmonger_MirrorFrame)
			ShowUIPanel(Gossipmonger_MirrorFrame)
		end
	end)
	QuestNPCModel:HookScript("OnHide", resetOffset)

	-- Adjust position at bank when Reagent tab is shown:
	ReagentBankFrame:HookScript("OnShow", function()
		--SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", (xoffset or 0) + ReagentBankFrame:GetWidth() - 70)
		SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", (xoffset or 0) + 316) -- using GetWidth wasn't reliable since it got the old bank frame's width, and sometimes got the new width if addons that switch directly to the reagent tab are used (like Reagent Banker)
		if (Gossipmonger_MirrorFrame:IsVisible()) then
			HideUIPanel(Gossipmonger_MirrorFrame)
			ShowUIPanel(Gossipmonger_MirrorFrame)
		end
	end)
	ReagentBankFrame:HookScript("OnHide", resetOffset)
  end

  Init = nil;
end

local function skipModifierKeyHeld()
  local key = Gossipmonger_Settings.greetingModifierKey
  if (key == 0) then
    return IsAltKeyDown()
  elseif (key == 1) then
    return IsControlKeyDown()
  elseif (key == 2) then
    return IsShiftKeyDown()
  end
  -- If key is set to None (3), it always returns nil.
end


local function buildOptionsTable(tab, isQuest, argsEach, ...)
  local list = tab.list and wipe(tab.list) or {}
  wipe(tab)
  local key2 = isQuest and "level" or "type"
  local count = 0
  for i = 1, select("#", ...), argsEach do
    count = count + 1
    local a, b, c = select(i, ...)
    tab["title"..count], list[#list+1] = a, a
    tab[key2..count], list[#list+1] = b, b
    if (isQuest) then
      c = not not c
      tab["isLowLevel"..count], list[#list+1] = c, c
    end
  end
  tab.list = list
end

local AvailableQuests, ActiveQuests, GossipOptions
local FILTERS = {}

local function findFilter(arg)
  if (type(arg) == "function") then
    for i,filter in ipairs(FILTERS) do
      if (filter.func == arg) then  return filter, i;  end
    end
  elseif (type(arg) == "string") then
    for i,filter in ipairs(FILTERS) do
      if (filter.name == arg) then  return filter, i;  end
    end
  end
end

local function callFilter(filter, mobID, NPCName, text, numAvailable, available, numActive, active, numGossip, gossip)
  if (filter.allowMulti or (numAvailable == 0 and numActive == 0 and numGossip == 1)) then
    if (filter.ignoreForce or not ForceGossip()) then
      return filter.func(mobID, NPCName, text, numAvailable, available, numActive, active, numGossip, gossip)
    end
  end
end

local OptionClicked_type, OptionClicked_id, OptionClicked_expect_numOptions, OptionClicked_expect_title
local function OptionClickedFollowup(mobID, NPCName, text, numAvailable, available, numActive, active, numGossip, gossip)
  --print("OptionClickedFollowup", numAvailable, numActive, numGossip, "=?", OptionClicked_expect_numOptions)
  if (numAvailable + numActive + numGossip == OptionClicked_expect_numOptions) then
    local tab
    if (OptionClicked_type == 1) then
      tab = available
    elseif (OptionClicked_type == 2) then
      tab = active
    elseif (OptionClicked_type == 3) then
      tab = gossip
    else
      return;
    end
    if (tab["title"..OptionClicked_id] == OptionClicked_expect_title) then
      return OptionClicked_type, OptionClicked_id, nil, Gossipmonger_Settings.showGossipMirrorOnQuestJump
    end
  end
end


-- GLOBAL FUNCTIONS
-------------------------

function Gossipmonger.SetFilter(func, name, desc, priority, allowMulti, ignoreForce)
-- Priority: Larger numbers go first. The default filter (which mimics standard WoW behavior) is set at priority 0.
-- Use nil for allowMulti or ignoreForce to default to false if the filter didn't already exist or stay at whatever it is currently
-- set to if it does exist. Use nil for name, desc, and/or priority to use the current value(s). If the filter didn't
-- already exist, a default name (e.g. "Unnamed Filter"), a blank description, and a priority of 1 are used.
  assert(type(func) == "function", "Usage: Gossipmonger.SetFilter(func[, name, desc, priority, allowMulti, ignoreForce]); func must be a function.")
  local tab, tabpos = findFilter(func)  -- See if filter is already used.

  local addtab
  if (tab) then
    name, desc = type(name) == "string" and name or tab.name, type(desc) == "string" and desc or tab.desc
    priority = type(priority) == "number" and priority or tab.priority
	if (allowMulti == nil) then  allowMulti = tab.allowMulti;  else  allowMulti = not not allowMulti;  end
	if (ignoreForce == nil) then  ignoreForce = tab.ignoreForce;  else  ignoreForce = not not ignoreForce;  end
    if (priority ~= tab.priority) then
      tremove(FILTERS, tabpos)
      addtab = true
    end
  else
    assert(priority == nil or type(priority) == "number", "Usage: Gossipmonger.SetFilter(func[, name, desc, priority, allowMulti, ignoreForce]); priority must be nil or a number.")
    name = type(name) == "string" and name or L.FILTER_UNNAMED
    desc = type(desc) == "string" and desc or nil
    priority = priority or 50
	allowMulti = not not allowMulti
	ignoreForce = not not ignoreForce
    tab, addtab = {}, true
	tab.enabled = false  -- New filters default to disabled.
  end

  tab.func, tab.name, tab.desc, tab.priority = func, name, desc, priority
  tab.allowMulti, tab.ignoreForce = allowMulti, ignoreForce

  if (addtab) then
    -- Find a suitable spot for a filter of this priority:
    for i,filter in ipairs(FILTERS) do
      if (filter.priority <= priority) then
        tinsert(FILTERS, i, tab)
        return;
      end
    end
    tinsert(FILTERS, tab)
  end
end

function Gossipmonger.GetFilterInfo(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then  return tab.func, tab.name, tab.desc, tab.priority, tab.allowMulti, tab.ignoreForce, tab.enabled;  end
end

function Gossipmonger.EnableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one (but only one of them will be affected).
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = true
    return true
  end
end

function Gossipmonger.DisableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one (but only one of them will be affected).
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = false
    return true
  end
end

function Gossipmonger.OnLoad()
  tinsert(UISpecialFrames, "Gossipmonger_MirrorFrame");
  --UIPanelWindows.Gossipmonger_MirrorFrame = { area = "right", pushable = 8, whileDead = 1, yoffset = -10 };
  --UIPanelWindows.Gossipmonger_MirrorFrame = { area = "right", pushable = 6, whileDead = 1, yoffset = -10 };
  UIPanelWindows.Gossipmonger_MirrorFrame = { area = "left", pushable = 6, whileDead = 1, yoffset = -10 };
  -- pushable had to be 6; 1-5 were too small, apparently; the bank window was to the Bubble's right. Unfortunate, as now the Character panel sits between the vendor window and the bubble.
  -- Could maybe make it dynamic, based on what we expect will appear, but that might not be worth the hassle.
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Init);
  Gossipmonger_MirrorFrame:RegisterEvent("VARIABLES_LOADED");
  Gossipmonger.OnLoad = nil
end

function Gossipmonger.OnEvent(self, event, arg1, ...)
  if (event == "BANKFRAME_CLOSED" or event == "BATTLEFIELDS_CLOSED" or event == "MERCHANT_CLOSED" or
      event == "TAXIMAP_CLOSED" or event == "TRAINER_CLOSED" or event == "QUEST_FINISHED" or event == "GOSSIP_CLOSED" or
      event == "AUCTION_HOUSE_CLOSED" or event == "SHIPMENT_CRAFTER_CLOSED" or event == "GARRISON_RECRUITMENT_NPC_CLOSED" or
	  event == "GARRISON_RECRUIT_FOLLOWER_RESULT" or event == "GARRISON_TRADESKILL_NPC_CLOSED") then
    if (event == "QUEST_FINISHED" and IgnoreQuestFinished and (IgnoreQuestFinished == true or GetTitleText() == IgnoreQuestFinished)) then
	  if (UnitExists("npc")) then  --if (QuestFrame:IsVisible() or GossipFrame:IsVisible()) then
	    IgnoreQuestFinished = nil
	    return;
	  end
    end
    BreakMirror()
  elseif (event == "GOSSIP_CONFIRM") then
    --print(event, ...)
    Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
  elseif (CurrentNPC == UnitName("npc")) then
    -- pre-BFA: if (event == "TAXIMAP_OPENED" and not TaxiFrame_ShouldShowOldStyle()) then
	if (event == "TAXIMAP_OPENED" and arg1 ~= Enum.UIMapSystem.Taxi) then
      -- New taximap style (from Legion):
	  -- Normally, does not support another panel like Gossip Bubble being open. If we open it now, then the bubble is opened first and the
	  -- flight path map doesn't display. If we delay opening it until after the flight path map is shown, then the map will be closed
	  -- once we show the bubble. However, we can open the bubble in a special way to get it to work well enough:
	  ShowMirror(FlightMapFrame)
	else
      if (event == "QUEST_ACCEPTED" or event =="QUEST_TURNED_IN") then
	    if (not QuestFrameGreetingPanel:IsVisible()) then  BreakMirror();  end -- Looking at the last quest on this NPC after completing/accepting another.
	    return;
	  end
	  if ((event == "GARRISON_RECRUITMENT_NPC_OPENED" or event == "GARRISON_RECRUITMENT_READY") and GarrisonRecruitSelectFrame and GarrisonRecruitSelectFrame:IsShown()) then
	    ShowMirror(GarrisonRecruitSelectFrame)
	  else
        ShowMirror()
	  end
	  if (event == "QUEST_DETAIL") then  -- Looking at another quest, perhaps because we just picked one up, so watch for that so we can remove the bubble when this was the last quest:
	    Gossipmonger_MirrorFrame:RegisterEvent("QUEST_ACCEPTED")
	    Gossipmonger_MirrorFrame:RegisterEvent("QUEST_TURNED_IN")
	  end
	end
  else
    BreakMirror()
  end
end

function Gossipmonger.GoodbyeClick()
  IgnoreQuestFinished = nil
  if (MerchantFrame:IsShown()) then
    HideUIPanel(MerchantFrame)
  elseif (BankFrame:IsShown()) then
    HideUIPanel(BankFrame)
  elseif (GossipFrame:IsShown()) then
    HideUIPanel(GossipFrame)
  elseif (QuestFrame:IsShown()) then
    HideUIPanel(QuestFrame)
  --[[
  elseif (BattlefieldFrame:IsShown()) then
    HideUIPanel(BattlefieldFrame)
  elseif (ArenaFrame:IsShown()) then
    HideUIPanel(ArenaFrame)
  --]]
  elseif (TaxiFrame:IsShown()) then
    HideUIPanel(TaxiFrame)
  elseif (FlightMapFrame and FlightMapFrame:IsShown()) then
    HideUIPanel(FlightMapFrame)
  elseif (PVEFrame:IsShown()) then -- Despite the name, this includes the PVP Queue.
    HideUIPanel(PVEFrame)
  elseif (ClassTrainerFrame and ClassTrainerFrame:IsShown()) then
    HideUIPanel(ClassTrainerFrame)
  elseif (AuctionFrame and AuctionFrame:IsShown()) then
    HideUIPanel(AuctionFrame)
  elseif (GarrisonCapacitiveDisplayFrame and GarrisonCapacitiveDisplayFrame:IsShown()) then
    HideUIPanel(GarrisonCapacitiveDisplayFrame)
  elseif (GarrisonRecruiterFrame and GarrisonRecruiterFrame:IsShown()) then
    HideUIPanel(GarrisonRecruiterFrame)
  elseif (GarrisonRecruitSelectFrame and GarrisonRecruitSelectFrame:IsShown()) then
    HideUIPanel(GarrisonRecruitSelectFrame)
  elseif (OrderHallTalentFrame and OrderHallTalentFrame:IsShown()) then
    HideUIPanel(OrderHallTalentFrame)

  elseif (CurrentNPC_gossip == "binder") then  -- Innkeeper
    local dialog = StaticPopup_FindVisible("CONFIRM_BINDER");
    if (dialog) then
      StaticPopup_OnClick(dialog, 2)   -- Call as if user pressed cancel button
    end
  elseif (CurrentNPC_gossip == "healer") then  -- Spirit Healer
    for k,v in ipairs(SPIRIT_HEALER_KEYS) do
      local dialog = StaticPopup_FindVisible(v);
      if (dialog) then
        StaticPopup_OnClick(dialog, 2)   -- Call as if user pressed cancel button
		break;
      end
    end

  elseif (TradeSkillFrame and TradeSkillFrame:IsShown()) then --Could possibly be from the player opening it (rather than a garrison NPC), so make sure this one's last.
    HideUIPanel(TradeSkillFrame)
  end

  -- Failsafe:  (normally, closing the above will close out the mirror frame and break the connection)
  --BreakMirror() -- comment out until we find reproduceable steps showing this is really needed with the changes made to the rest of the code
end

function Gossipmonger.OptionClick(self)
  if (not SafeMirrorOptionsClick or self.selected) then  return;  end
  if (QuestFrame:IsShown()) then
    local optionType, typeid = self.type
    if (optionType == "Available") then
      typeid = 1
    elseif (optionType == "Active") then
      typeid = 2
    elseif (optionType == "Gossip") then
      typeid = 3
    elseif (self.QuestGreeting) then
      local icon = _G[self.baseButton.."QuestIcon"]
      if (icon:GetTexture() == "Interface\\GossipFrame\\ActiveQuestIcon") then
        typeid = 2
      else
        typeid = 1
      end
    else
      return;
    end

	IgnoreQuestFinished = true  -- Needed so using the hopping feature to select a vendor option will show the mirror once the vendor frame opens.
	-- (Although you can't make another hop from the vendor frame, showing it then can be good since the NPC's gossip text may have changed.)

    OptionClicked_type, OptionClicked_id = typeid, self:GetID()
    OptionClicked_expect_numOptions = Gossipmonger_MirrorFrame.numOptions
    OptionClicked_expect_title = self.title
    GossipNPC = UnitName("npc")
    FollowupFunction = OptionClickedFollowup
    FollowupFunction_filtername = "<QUESTJUMPER>"
	--print("set",GossipNPC,FollowupFunction)

    SafeMirrorOptionsClick = nil
    DeclineQuest();
    --PlaySound("igQuestCancel"); -- redundant? even in gossip quest frame?
    --HideUIPanel(Gossipmonger_MirrorFrame)  -- Hide the mirror (but don't "break" it)

    if (not GossipFrame:IsShown() and not QuestFrame:IsShown()) then
      -- In some cases, DeclineQuest() will close the quest frame, making quest-jumping inoperable.
      -- Not sure exactly how to detect this is going to happen before it does. So far it seems to happen with the QuestFrame and not the GossipFrame,
      -- but there are QuestFrame usages where it doesn't happen. All the QuestFrame stuff seemed just fine, including jumping, until I found one NPC
	  -- where it wouldn't do it. If that happens, then don't save jump data. Otherwise, it will try to go to the chosen quest when you talk to the
      -- NPC next, which maybe is convenient *sometimes* but it can be confusing and it also seems to break other things about the Gossip Bubble.
      OptionClicked_type, OptionClicked_id, OptionClicked_expect_numOptions, OptionClicked_expect_title = nil, nil, nil, nil
	  GossipNPC, FollowupFunction, FollowupFunction_filtername = nil, nil, nil
    end
  end
end

--[[ DEBUG: Force the "Gossip Bubble" (AKA Mirror Frame) to show:
function Gossipmonger.ShowBubble(delay, name, text, selected, gossipType)
  name = name or UnitName("npc")
  text = text and text:gsub("||n","|n") or GetGossipText()
  showGossipMirror(delay, name, text, selected, gossipType)
end
--]]

Gossipmonger.PrepGossipOptionDisplay = prepGossipOptionDisplay

Gossipmonger.ShowMirror = ShowMirror


-- HOOK
-------------------------

local orig_GossipFrame_OnEvent = GossipFrame_OnEvent

local function filtererror(filter, msg, ...)
  orig_GossipFrame_OnEvent(...)
  local name = filter and '"'..filter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
  FollowupFunction = nil
  error("Invalid return from filter ("..name.."): "..msg, 2)
end

local function ShowGossipFrame(self)
  -- Show default gossip frame (see GossipFrame_OnEvent in FrameXML\GossipFrame.lua):
  if ( not GossipFrame:IsShown() ) then
    ShowUIPanel(self);
    if ( not self:IsShown() ) then
      CloseGossip();
      return;
    end
  end
  NPCFriendshipStatusBar_Update(self);
  GossipFrameUpdate();

  return true
end

GossipFrame_OnEvent = function(self, event, ...)
  if (event == "GOSSIP_SHOW") then
    --BreakMirror() -- needed here? It was found to be needed for QuestFrameGreetingPanel:OnShow hook.

    local name = UnitName("npc")
    local text = GetGossipText()
    local numAvailable, numActive, numGossip = GetNumGossipAvailableQuests(), GetNumGossipActiveQuests(), GetNumGossipOptions()

    local validFollowup, showMirrorIfNoFollowupResult
    if (GossipNPC == name or BlockNextFilter == name) then
    -- Dialog continued from auto-selected gossip option or player-canceled quest display:
      if (FollowupFunction and not BlockNextFilter) then
        validFollowup = FollowupFunction
        showMirrorIfNoFollowupResult = ShowMirrorOnGossip
      else
        if (not ShowGossipFrame(self)) then
          BreakMirror()
        elseif (BlockNextFilter) then
        -- This is the result of the player declining a quest or canceling the turn-in of a quest:
          if (numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
          -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
          -- the Gossipmonger "quest hanger" feature:
            showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
          else
            Gossipmonger_MirrorFrame.selected = nil
            Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
            Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
            Gossipmonger_MirrorFrame.NPCName = name
          end
        elseif (ShowMirrorOnGossip) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        end
        GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
        return;
      end
    end
    GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil

    -- New dialog or using a followup function:
    local skipdefault, keyheld = Gossipmonger_Settings.skipIsDefault, skipModifierKeyHeld()
    if ( (validFollowup or (not keyheld and skipdefault) or (keyheld and not skipdefault)) and
         (numAvailable > 0 or numActive > 0 or numGossip > 0) ) then
      AvailableQuests, ActiveQuests, GossipOptions = AvailableQuests or {}, ActiveQuests or {}, GossipOptions or {}
      buildOptionsTable(AvailableQuests, true, 7, GetGossipAvailableQuests())
      buildOptionsTable(ActiveQuests, true, 6, GetGossipActiveQuests())
      buildOptionsTable(GossipOptions, false, 2, GetGossipOptions())
      local seltype, seloption, followup, override
      local thefilter

      if (validFollowup) then
        seltype, seloption, followup, override =
          validFollowup(getMobID("npc"), name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions, true)
      else
	    local mobID = getMobID("npc")
        for i,filter in ipairs(FILTERS) do
          if (filter.enabled) then
            seltype, seloption, followup, override =
              callFilter(filter, mobID, name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions)
            if (seltype and (seloption or seltype == 0)) then
              thefilter = filter
              break
            end
          end
        end
      end

      if (seltype and seloption and seltype ~= 0 and seloption ~= 0) then  -- Filter with results found:
        -- If seltype is 0, this section is skipped, causing the default gossip frame to be shown.
        if (type(seloption) ~= "number") then
          filtererror(thefilter, "Second return (option) must be a number.", self, event, ...)
          return;
        end
        local numOptions, SelectOption, gossipType
        if (seltype == 1) then
          numOptions, SelectOption = numAvailable, SelectGossipAvailableQuest
        elseif (seltype == 2) then
          numOptions, SelectOption = numActive, SelectGossipActiveQuest
        elseif (seltype == 3) then
          numOptions, SelectOption = numGossip, SelectGossipOption
          gossipType = GossipOptions["type"..seloption]
        else
          filtererror(thefilter, "First return (type) must be an integer between 0 and 3.", self, event, ...)
          return;
        end
        if (seloption < 1 or seloption > numOptions) then
          filtererror(thefilter, "Option of that type ("..seltype..") and index ("..seloption..") not found.", self, event, ...)
          return;
        end

        checkInform(override, validFollowup, gossipType, name)

        -- If override is true, the mirror is forced to show. If it is false (not nil), it is forced not to.
        if (override == false) then
          BreakMirror()
          if (seltype ~= 3) then
            -- Prepare in case the "quest hanger" icon is clicked:
            showGossipMirror(true, name, text, index, gossipType, true, nil, numAvailable + numActive)
          end
        else
          local index = seloption
          if (seltype > 1) then  index = index + numAvailable;  end
          if (seltype > 2) then  index = index + numActive;  end
          if (override or Gossipmonger_Settings.showGossipMirrorOnSkip) then
            showGossipMirror(Gossipmonger_Settings.delayShowingMirror, name, text, index, gossipType, nil, (seltype ~= 3), numAvailable + numActive)
            ShowMirrorOnGossip = true
          elseif (seltype ~= 3) then
            -- Prepare in case the "quest hanger" icon is clicked:
            showGossipMirror(true, name, text, index, gossipType, true, 1, numAvailable + numActive)
          end
        end

        if (gossipType == "gossip") then
          GossipNPC = name
          if (type(followup) == "function") then
            FollowupFunction = followup
            FollowupFunction_filtername = thefilter.name
          end
        end
        --SelectOption(seloption)
		if (not validFollowup or ShowGossipFrame(self)) then
		  Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CONFIRM")
		  SelectOption(seloption)
		end
        return;

      elseif (showMirrorIfNoFollowupResult) then
        if (ShowGossipFrame(self)) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        else
          BreakMirror()
        end
        return;
      end
    end

    local shown = ShowGossipFrame(self)

    if (Gossipmonger_MirrorFrame:IsShown()) then
    -- Fix the odd case where the mirror is sometimes left up after it was shown without a delay and the
    -- player ran away from the NPC before the selected option's result appears. (still relevant in latest version?)
      BreakMirror()
    end

    if (shown and numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
    -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
    -- the Gossipmonger "quest hanger" feature:
      showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
    else
      Gossipmonger_MirrorFrame.selected = nil
      Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
      Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
      Gossipmonger_MirrorFrame.NPCName = name
    end

  else
   if (event == "GOSSIP_CLOSED") then
      GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
   end
    orig_GossipFrame_OnEvent(self, event, ...)
  end
end

GossipFrame:SetScript("OnEvent", GossipFrame_OnEvent) -- Need to do this as of WoW 3.1 due to XML's function=<x> usage.


-- DETECT POPUP CLOSING SO WE CAN CLOSE RELATED GOSSIP BUBBLES:

do
  local function hookPopupOnHide(key)
    if (StaticPopupDialogs[key].OnHide) then
      hooksecurefunc(StaticPopupDialogs[key], "OnHide", BreakMirror)
    else
      StaticPopupDialogs[key].OnHide = BreakMirror
    end
  end

  -- Innkeeper
  hookPopupOnHide("CONFIRM_BINDER")

  -- Spirit Healer
  for k,v in ipairs(SPIRIT_HEALER_KEYS) do
    hookPopupOnHide(v)
  end
end

--[[
local orig_onHide = StaticPopupDialogs["CONFIRM_BINDER"].OnHide;
StaticPopupDialogs["CONFIRM_BINDER"].OnHide = function(...)
  BreakMirror();
  if (orig_onHide) then  orig_onHide(...)  end
end
--]]



-- This function helps prevent Gossipmonger from running its filters when options appear due to the player hitting
-- Cancel buttons (e.g. "Cancel", "Decline", "Ignore", or "Unignore") to decline an offered quest or cancel turning in a quest.
local function QuestDeclined_DoNotFilter()
  BlockNextFilter = UnitName("npc")
end

--hooksecurefunc("QuestGoodbyeButton_OnClick", QuestDeclined_DoNotFilter)
--hooksecurefunc("QuestRewardCancelButton_OnClick", QuestDeclined_DoNotFilter)
--hooksecurefunc("QuestDetailDeclineButton_OnClick", QuestDeclined_DoNotFilter)

QuestFrameGoodbyeButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- "Cancel" when looking at in-progress quest
QuestFrameDeclineButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- "Decline" offered quest
--QuestFrameCancelButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button no longer there.
--QuestFrameDetailPanel.IgnoreButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button closes window so shouldn't have this hook.
--QuestFrameDetailPanel.UnignoreButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button closes window so shouldn't have this hook.
-- In fact, the other two buttons now close the window, so is this hook needed at all any more?

-- This function prevents the mirror from disappearing if it was visible when the player hits the "Continue"
-- button (for completing a quest).
local function QuestFrameShiftFromProgressToRewards()
  if (Gossipmonger_MirrorFrame:IsShown()) then
    IgnoreQuestFinished = GetTitleText()
  else
    IgnoreQuestFinished = nil
  end
end

--hooksecurefunc("QuestProgressCompleteButton_OnClick", QuestFrameShiftFromProgressToRewards)
QuestFrameCompleteButton:HookScript("OnClick", QuestFrameShiftFromProgressToRewards)


-- Handle the quest frame, which is used by a few NPCs which have no gossip (though many (most?) without gossip still use the gossip frame):
QuestFrameGreetingPanel:HookScript("OnShow", function()
  --print("QuestFrameGreetingPanel")
  BreakMirror()

  local name = UnitName("npc")
  local text = GetGreetingText()
  local numAvailable, numActive = GetNumAvailableQuests(), GetNumActiveQuests()

  local validFollowup = GossipNPC == name and FollowupFunction == OptionClickedFollowup and FollowupFunction
  GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil

  if (validFollowup) then  -- Quest-jumping button clicked:
    AvailableQuests = AvailableQuests and wipe(AvailableQuests) or {}
    ActiveQuests = ActiveQuests and wipe (ActiveQuests) or {}
    GossipOptions = GossipOptions and wipe(GossipOptions) or {}

    local list = ActiveQuests.list and wipe(ActiveQuests.list) or {}
    local a, b, c
    for i = 1, numActive do
      a, b, c = GetActiveTitle(i), GetActiveLevel(i), IsActiveQuestTrivial(i)
      ActiveQuests["title"..i], list[#list+1] = a, a
      ActiveQuests["level"..i], list[#list+1] = b, b
      ActiveQuests["isLowLevel"..i], list[#list+1] = c, c
    end
    ActiveQuests.list = list

    list = AvailableQuests.list and wipe(AvailableQuests.list) or {}
    for i = 1, numAvailable do
      a, b, c = GetAvailableTitle(i), GetAvailableLevel(i), IsAvailableQuestTrivial(i)
      AvailableQuests["title"..i], list[#list+1] = a, a
      AvailableQuests["level"..i], list[#list+1] = b, b
      AvailableQuests["isLowLevel"..i], list[#list+1] = c, c
    end
    AvailableQuests.list = list

    local seltype, seloption, followup, override =
      validFollowup(getMobID("npc"), name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, 0, GossipOptions, true)

    if (seltype and seloption and seltype ~= 0 and seloption ~= 0) then
      -- If seltype is 0, this section is skipped, causing the default frame to be shown.
      if (type(seloption) ~= "number") then
        --filtererror(thefilter, "Second return (option) must be a number.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): ".."Second return (option) must be a number.")
        return;
      end
      local numOptions, SelectOption, gossipType
      if (seltype == 1) then
        numOptions, SelectOption = numAvailable, SelectAvailableQuest
      elseif (seltype == 2) then
        numOptions, SelectOption = numActive, SelectActiveQuest
      elseif (seltype == 3) then
        numOptions, SelectOption = numGossip, SelectGossipOption
        gossipType = GossipOptions["type"..seloption]
      else
        --filtererror(thefilter, "First return (type) must be an integer between 0 and 3.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): ".."First return (type) must be an integer between 0 and 3.")
        return;
      end
      if (seloption < 1 or seloption > numOptions) then
        --filtererror(thefilter, "Option of that type ("..seltype..") and index ("..seloption..") not found.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): "..("Option of that type ("..seltype..") and index ("..seloption..") not found."))
        return;
      end

      checkInform(override, validFollowup, gossipType, name)

      -- If override is true, the mirror is forced to show. If it is false (not nil), it is forced not to.
      if (override == false) then
        BreakMirror()
        if (seltype ~= 3) then
          -- Prepare in case the "quest hanger" icon is clicked:
          showGossipMirror(true, name, text, index, gossipType, true, nil, numAvailable + numActive, true)
        end
      else
        local index = seloption
        if (seltype == 1) then  index = index + numAvailable;  end
          -- Note that this is different from the normal method because active quests are displayed before available
          -- quests here (the reverse of normal).
        if (override or Gossipmonger_Settings.showGossipMirrorOnSkip) then
          showGossipMirror(Gossipmonger_Settings.delayShowingMirror, name, text, index, gossipType, nil, (seltype ~= 3), numAvailable + numActive, true)
          ShowMirrorOnGossip = true
        elseif (seltype ~= 3) then
          -- Prepare in case the "quest hanger" icon is clicked:
          showGossipMirror(true, name, text, index, gossipType, true, 1, numAvailable + numActive, true)
        end
      end

      if (gossipType == "gossip") then
        GossipNPC = name
        if (type(followup) == "function") then
          FollowupFunction = followup
          FollowupFunction_filtername = thefilter.name
        end
      end
      SelectOption(seloption)
      --[[ NEED TO TEST whether the line above alone is enough. It used to be enough for the GossipFrame version, but it stopped working so the below was
           added to that function. But it's likely that if the line above isn't enough, the code below won't be what we need, either, since it's for the
           gossip frame, not QuestFrameGreetingPanel. Some adaptation of it could work, though (e.g. forcing QuestFrameGreetingPanel to show/update?).
	  if (not validFollowup or ShowGossipFrame(self)) then
	    SelectOption(seloption)
	  end
	  --]]
      return;
    end
  end

  if (numAvailable + numActive >= 1 and numAvailable + numActive > 1) then
  -- If there are multiple quests, prepare the mirror in case it's used by the Gossipmonger "quest hanger" feature:
    showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive, true)
	Gossipmonger_MirrorFrame.NPCName = name
  else
    Gossipmonger_MirrorFrame.selected = nil
    Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive
    Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
    Gossipmonger_MirrorFrame.NPCName = name
  end
end)


-- TEST LOCATION: Refuge Pointe. Skuerto (NPC) and Wanted Board (object).


